import logging
from copy import deepcopy
from functools import wraps
from io import BytesIO
from pathlib import Path
from threading import Lock as ThreadLock
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Optional,
    TYPE_CHECKING,
    Tuple,
    TypeVar,
    Union,
)

from aiofiles.tempfile import SpooledTemporaryFile
from aiohttp.typedefs import (
    LooseHeaders,
    StrOrURL,
)
from arkowrapper import ArkoWrapper
from typing_extensions import ParamSpec

from async_pixiv.client._section import SectionType
from async_pixiv.error import (
    LoginError,
    OauthError,
)
from async_pixiv.model.model import User
from async_pixiv.utils.model import Net
from async_pixiv.utils.typed import RequestMethod

if TYPE_CHECKING:
    from async_pixiv.client._section import (
        USER,
        ILLUST,
    )
    from aiohttp import (
        ClientResponse,
    )

__all__ = ['PixivClient']

_REDIRECT_URI = "https://app-api.pixiv.net/web/v1/users/auth/pixiv/callback"
_LOGIN_URL = "https://app-api.pixiv.net/web/v1/login"
_LOGIN_VERIFY = "https://accounts.pixiv.net/ajax/login"
_AUTH_TOKEN_URL = "https://oauth.secure.pixiv.net/auth/token"

P = ParamSpec('P')
R = TypeVar('R')
T = TypeVar('T')
logger = logging.getLogger('async_pixiv.client')


# noinspection PyPep8Naming
class PixivClient(Net):
    _instances: ClassVar[List["PixivClient"]] = []

    @classmethod
    def get_client(cls) -> "PixivClient":
        if cls._instances:
            return cls._instances[0]
        else:
            raise ValueError("请先实列化一个 PixivClient ")

    _lock: ThreadLock = ThreadLock()

    class Config(NamedTuple):
        client_id: str
        client_secret: str
        user_agent: str

    _config: Config
    _request_headers: Dict[str, Any]

    access_token: Optional[str] = None
    refresh_token: Optional[str] = None
    account: Optional[User] = None
    _sections: Dict[str, SectionType] = {}

    @property
    def USER(self) -> "USER":
        with self._lock:
            if self._sections.get('user', None) is None:
                from async_pixiv.client._section import USER
                self._sections['user'] = USER(self)
        return self._sections['user']

    @property
    def ILLUST(self) -> "ILLUST":
        with self._lock:
            if self._sections.get('illust', None) is None:
                from async_pixiv.client._section import ILLUST
                self._sections['illust'] = ILLUST(self)
        return self._sections['illust']

    @property
    def is_logged(self) -> bool:
        return self.refresh_token is not None

    def need_logged(self, func: Callable[[P], R]) -> Callable[[P], R]:
        @wraps(func)
        def wrapper():
            if not self.is_logged:
                raise OauthError("Please login.")
            return func

        return wrapper

    def __init__(
            self, *,
            limit: int = 30,
            timeout: float = 10,
            proxy: Optional[StrOrURL] = None,
            trust_env: bool = False,
    ):
        super().__init__(
            limit=limit,
            timeout=timeout,
            proxy=proxy,
            trust_env=trust_env
        )
        config_path = (Path(__file__) / "../config").resolve()
        with open(config_path, encoding='utf-8') as file:
            self._config = PixivClient.Config(
                *ArkoWrapper(file.readlines()).map(lambda x: x.strip())
            )
        self._request_headers = {
            'App-OS': 'ios',
            'App-OS-Version': '12.2',
            'App-Version': '7.6.2',
            'user-agent': self._config.user_agent,
            'Referer': 'https://app-api.pixiv.net/'
        }
        PixivClient._instances.append(self)

    async def _request(
            self,
            method: RequestMethod,
            url: StrOrURL, *,
            params: Optional[Mapping[str, Any]] = None,
            headers: Optional[LooseHeaders] = None,
            data: Any = None,
    ) -> "ClientResponse":
        request_headers = deepcopy(self._request_headers)
        if headers is not None:
            for key in headers:
                request_headers[key.lower().title()] = headers[key]
        if self.access_token is not None:
            request_headers.update({
                'Authorization': f"Bearer {self.access_token}"
            })
        return await super(PixivClient, self)._request(
            method, url, params=params, headers=request_headers, data=data
        )

    async def login_with_pwd(self, username: str, password: str) -> User:
        try:
            from playwright.async_api import async_playwright
        except ImportError:
            raise ImportError("Please install 'playwright'.")

        def oauth_pkce() -> Tuple[str, str]:
            from secrets import token_urlsafe
            from hashlib import sha256
            from base64 import urlsafe_b64encode
            verifier = token_urlsafe(32)
            challenge = urlsafe_b64encode(
                sha256(verifier.encode('ascii')).digest()
            ).rstrip(b'=').decode('ascii')
            return verifier, challenge

        async with async_playwright() as playwright:
            from urllib.parse import (
                parse_qs,
                urlparse,
                urlencode,
            )
            for p in self._proxies:
                if p.scheme not in ["https", 'wss']:
                    proxy = str(p)
                    break
            # noinspection PyTypeChecker
            browser = await playwright.chromium.launch(
                proxy={"server": proxy}
            )
            context = await browser.new_context()
            api_request_context = context.request
            page = await context.new_page()

            # 访问登录页面
            code_verifier, code_challenge = oauth_pkce()
            await page.goto(urlparse(_LOGIN_URL)._replace(
                query=urlencode({
                    "code_challenge": code_challenge,
                    "code_challenge_method": "S256",
                    "client": "pixiv-android",
                })
            ).geturl(), timeout=0)

            # 输入用户名与密码
            await page.locator(".degQSE").fill(username)
            await page.locator(".hfoSmp").fill(password)

            # 点击登录按钮
            # noinspection SpellCheckingInspection
            await page.locator("button.sc-bdnxRM:nth-child(5)").click()

            # 验证登录
            # noinspection PyBroadException
            try:
                async with page.expect_response(f"{_LOGIN_VERIFY}*") as future:
                    response = await (await future.value).json()
                raise LoginError(
                    f"登录错误，请检查登录的用户名和密码是否正确：{response['body']['errors']}"
                )
            except Exception as e:
                if not isinstance(e, LoginError):
                    logger.debug("登录成功，正尝试获取 token")

            # 获取code
            async with page.expect_request(f"{_REDIRECT_URI}*") as request:
                url = urlparse((await request.value).url)
            code = parse_qs(url.query)['code'][0]

            # 获取token
            response = await api_request_context.post(
                _AUTH_TOKEN_URL,
                form={
                    "client_id": self._config.client_id,
                    "client_secret": self._config.client_secret,
                    "code": code,
                    "code_verifier": code_verifier,
                    "grant_type": "authorization_code",
                    "include_policy": "true",
                    "redirect_uri": _REDIRECT_URI,
                },
                headers={
                    'Accept-Encoding': 'gzip, deflate',
                    "Content-Type": 'application/x-www-form-urlencoded',
                    "User-Agent": self._config.user_agent,
                    'Host': 'oauth.secure.pixiv.net'
                },
                timeout=0
            )
            data = await response.json()
            await browser.close()
        self.access_token = data["access_token"]
        self.refresh_token = data["refresh_token"]
        self.account = User.parse_obj(data['user'])
        return self.account

    async def login_with_token(self, token: str) -> User:
        response = await self.post(
            _AUTH_TOKEN_URL, data={
                "client_id": self._config.client_id,
                "client_secret": self._config.client_secret,
                "grant_type": "refresh_token",
                "include_policy": "true",
                "refresh_token": token,
            },
            headers={"User-Agent": self._config.user_agent}
        )
        data = await response.json()
        self.access_token = data["access_token"]
        self.refresh_token = data["refresh_token"]
        self.account = User.parse_obj(data['user'])
        return self.account

    async def login(
            self,
            username: Optional[str] = None,
            password: Optional[str] = None,
            token: Optional[str] = None
    ) -> User:
        if token is not None:
            return await self.login_with_token(token)
        else:
            return await self.login_with_pwd(username, password)

    async def download(
            self, url: StrOrURL, *,
            output: Optional[Union[str, Path, BytesIO]] = None
    ) -> Optional[bytes]:
        if output is None:
            out = True
            output = BytesIO()
        else:
            out = False
            output = Path(output).resolve().open('wb+')
        content = (await self.get(url)).content
        async with SpooledTemporaryFile(mode='wb+') as file:
            # noinspection PyProtectedMember
            file._file._file = output
            while True:
                data, is_end = await content.readchunk()
                await file.write(data)
                if is_end or not data:
                    break
            await file.seek(0)
            if out:
                return await file.read()



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Iterator,
    List,
    Optional,
    TYPE_CHECKING,
)

from pydantic import (
    Field,
    HttpUrl,
)
from typing_extensions import Self

from async_pixiv.model._base import PixivModel
from async_pixiv.model.artwork import (
    ArtWork,
    Comment,
)
from async_pixiv.model.user import (
    User,
    UserProfile,
    UserProfilePublicity,
    UserWorkSpace,
)

if TYPE_CHECKING:
    from async_pixiv.client import PixivClient


class SearchResult(ABC, PixivModel):
    next_url: Optional[HttpUrl]
    search_span_limit: Optional[int]

    @abstractmethod
    def __iter__(self) -> Iterator:
        pass

    async def next(self, client: Optional["PixivClient"] = None) -> Self:
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        data = await (await client.get(str(self.next_url))).json()
        return self.__class__.parse_obj(data)


class UserPreview(PixivModel):
    user: User
    illusts: List[ArtWork]
    is_muted: bool


class UserSearchResult(SearchResult):
    users: List[UserPreview] = Field(alias='user_previews')

    def __iter__(self) -> Iterator[UserPreview]:
        return iter(self.users)


class UserIllustsResult(PixivModel):
    illusts: List[ArtWork]
    next_url: Optional[HttpUrl]


class UserBookmarksIllustsResult(UserIllustsResult):
    pass


class UserDetailResult(PixivModel):
    user: User
    profile: UserProfile
    profile_publicity: UserProfilePublicity
    workspace: UserWorkSpace


class UserRelatedResult(PixivModel):
    users: List[UserPreview] = Field(alias='user_previews')

    def __iter__(self) -> Iterator[UserPreview]:
        return iter(self.users)


class IllustSearchResult(SearchResult):
    illusts: List[ArtWork]

    def __iter__(self) -> Iterator[ArtWork]:
        return iter(self.illusts)


class IllustDetailResult(PixivModel):
    illust: ArtWork


class IllustCommentResult(SearchResult):
    total: int = Field(alias='total_comments')
    comments: List[Comment]
    next_url: Optional[HttpUrl]
    access_comment: Optional[bool] = Field(alias='comment_access_control')

    def __iter__(self) -> Iterator[Comment]:
        return iter(self.comments)


class IllustRelatedResult(IllustSearchResult):
    pass



from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import (
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
    Union,
)

from pydantic import (
    Field,
    HttpUrl,
    validator,
)

from async_pixiv.model._base import PixivModel
from async_pixiv.model.other import ImageUrl
from async_pixiv.model.user import User

if TYPE_CHECKING:
    from async_pixiv.client import PixivClient
    from async_pixiv.model.result import (
        IllustDetailResult,
        IllustCommentResult,
        IllustRelatedResult,
    )

__all__ = [
    'Tag', 'TagTranslation',
    'ArtWork',
    'Comment',
]


class TagTranslation(PixivModel):
    zh: Optional[str]
    en: Optional[str]


class Tag(PixivModel):
    name: str
    translated_name: Optional[str]


class ArtWorkType(Enum):
    illust = 'illust'
    ugoira = 'ugoira'
    manga = 'manga'


# noinspection PyProtectedMember
class ArtWork(PixivModel):
    class Series(PixivModel):
        id: int
        title: str

    id: int
    title: str
    type: ArtWorkType
    image_urls: "ImageUrl"
    caption: str
    restrict: int
    user: "User"
    tags: List[Tag]
    tools: List[str]
    create_date: datetime
    page_count: int
    width: int
    height: int
    sanity_level: int
    x_restrict: int
    series: Optional[Series]
    meta_single_page: Dict[str, HttpUrl]
    meta_pages: List["ImageUrl"]
    total_view: int
    total_bookmarks: int
    is_bookmarked: bool
    visible: bool
    is_muted: bool
    total_comments: Optional[int]
    comment_access_control: Optional[int]

    @property
    def is_r18(self) -> bool:
        if len(self.tags) >= 1:
            return self.tags[0].name.title().replace('-', '') == 'R18'
        return False

    async def detail(
            self, client: Optional["PixivClient"] = None, *,
            for_ios: bool = True
    ) -> "IllustDetailResult":
        from async_pixiv.client._section import SearchFilter

        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()

        return await client.ILLUST.detail(
            self.id,
            filter=SearchFilter.ios if for_ios else SearchFilter.android
        )

    async def comments(
            self, client: Optional["PixivClient"] = None, *,
            offset: Optional[int] = None
    ) -> "IllustCommentResult":
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        return await client.ILLUST.comments(self.id, offset=offset)

    async def related(
            self, client: Optional["PixivClient"] = None, *,
            for_ios: bool = True,
            offset: Optional[int] = None,
            seed_id: Optional[int] = None
    ) -> "IllustRelatedResult":
        from async_pixiv.client._section import SearchFilter
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        return await client.ILLUST.related(
            self.id, offset=offset, seed_ids=seed_id,
            filter=SearchFilter.ios if for_ios else SearchFilter.android
        )

    async def download(
            self, *,
            output: Optional[Union[str, Path]] = None,
            client: Optional["PixivClient"] = None
    ) -> Optional[bytes]:
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        url = str(self.image_urls.original or self.image_urls.large)
        return await client.download(url, output=output)


class Comment(PixivModel):
    id: int
    comment: str
    date: datetime
    user: User
    parent: Optional["Comment"] = Field(alias='parent_comment')

    @validator('parent', pre=True, always=True)
    def parent_check(cls, value):
        if value == {}:
            return None
        else:
            return value



from typing import (
    Optional,
    TYPE_CHECKING,
)

from pydantic import HttpUrl

from async_pixiv.model._base import PixivModel
from async_pixiv.model.other import ImageUrl

if TYPE_CHECKING:
    from async_pixiv.client import PixivClient
    from async_pixiv.model.result import (
        UserDetailResult,
        UserIllustsResult,
        UserBookmarksIllustsResult,
        UserRelatedResult,
    )


# noinspection PyProtectedMember
class User(PixivModel):
    id: int
    name: str
    account: str
    profile_image_urls: ImageUrl
    is_followed: Optional[bool]
    comment: Optional[str]

    async def detail(
            self, client: Optional["PixivClient"] = None, *,
            for_ios: bool = True
    ) -> "UserDetailResult":
        from async_pixiv.client._section import SearchFilter

        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()

        return await client.USER.detail(
            self.id,
            filter=SearchFilter.ios if for_ios else SearchFilter.android
        )

    async def illusts(
            self, client: Optional["PixivClient"] = None, *,
            for_ios: bool = True, offset: Optional[int] = None
    ) -> "UserIllustsResult":
        from async_pixiv.client._section import SearchFilter

        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        return await client.USER.illusts(
            self.id, offset=offset,
            filter=SearchFilter.ios if for_ios else SearchFilter.android,
        )

    async def bookmarks(
            self, client: Optional["PixivClient"] = None, *,
            for_ios: bool = True, tag: Optional[str] = None,
            max_bookmark_id: Optional[int] = None,
    ) -> "UserBookmarksIllustsResult":
        from async_pixiv.client._section import SearchFilter

        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        return await client.USER.bookmarks(
            self.id, tag=tag, max_bookmark_id=max_bookmark_id,
            filter=SearchFilter.ios if for_ios else SearchFilter.android,
        )

    async def related(
            self, client: Optional["PixivClient"] = None, *,
            for_ios: bool = True, offset: Optional[int] = None
    ) -> "UserRelatedResult":
        from async_pixiv.client._section import SearchFilter

        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        return await client.USER.related(
            self.id, offset=offset,
            filter=SearchFilter.ios if for_ios else SearchFilter.android,
        )


class UserProfile(PixivModel):
    webpage: Optional[HttpUrl]
    gender: str
    birth: str
    birth_day: str
    birth_year: int
    region: str
    address_id: int
    country_code: str
    job: str
    job_id: int
    total_follow_users: int
    # noinspection SpellCheckingInspection
    total_mypixiv_users: int
    total_illusts: int
    total_manga: int
    total_novels: int
    total_illust_bookmarks_public: int
    total_illust_series: int
    total_novel_series: int
    background_image_url: Optional[HttpUrl]
    twitter_account: Optional[str]
    twitter_url: Optional[HttpUrl]
    pawoo_url: Optional[HttpUrl]
    is_premium: bool
    is_using_custom_profile_image: bool


class UserProfilePublicity(PixivModel):
    gender: str
    region: str
    birth_day: str
    birth_year: str
    job: str
    pawoo: bool


class UserWorkSpace(PixivModel):
    pc: Optional[str]
    monitor: Optional[str]
    tool: Optional[str]
    scanner: Optional[str]
    tablet: Optional[str]
    mouse: Optional[str]
    printer: Optional[str]
    desktop: Optional[str]
    music: Optional[str]
    desk: Optional[str]
    chair: Optional[str]
    comment: Optional[str]
    workspace_image_url: Optional[HttpUrl]



import asyncio
import sys
from asyncio import Lock
from enum import Enum
from typing import (
    Any,
    List,
    Mapping,
    NamedTuple,
    Optional,
    TYPE_CHECKING,
)

from aiohttp import (
    ClientSession,
    ClientTimeout,
    TCPConnector,
)
from aiohttp.typedefs import (
    LooseHeaders,
    StrOrURL,
)
from aiohttp_socks import ProxyConnector
from typing_extensions import Self
from yarl import URL

from async_pixiv.utils.func import proxies_from_env
from async_pixiv.utils.typed import RequestMethod

try:
    import regex as re
except ImportError:
    import re

try:
    import ujson as json
except ImportError:
    import json

try:
    from uvloop import EventLoopPolicy
except ImportError:
    if sys.platform.startswith('win'):
        from asyncio import WindowsProactorEventLoopPolicy as EventLoopPolicy
    else:
        from asyncio import DefaultEventLoopPolicy as EventLoopPolicy

if TYPE_CHECKING:
    from aiohttp import ClientResponse

__all__ = ['Net']

asyncio.set_event_loop_policy(EventLoopPolicy())


class Config(NamedTuple):
    client_id: str
    client_secret: str
    user_agent: str


class Net(object):
    __slots__ = (
        '_conn_kwargs', '_timeout', '_session', '_proxies', '_trust_env',
    )

    _limit: int
    _timeout: ClientTimeout
    _trust_env: bool
    _proxies: Optional[List[URL]]
    _session: Optional[ClientSession]

    _aio_lock: Lock = Lock()

    async def _init_session(self) -> ClientSession:
        async with self._aio_lock:
            if self._session is None or self._session.closed:
                proxy: Optional[URL] = None

                for p in self._proxies:
                    if p.scheme in ['socks5', 'socks4', 'http', 'ws']:
                        proxy = p
                        break

                if (
                        proxy is not None
                        and
                        proxy.scheme in ['socks5', 'socks4', 'http']
                ):
                    connector = ProxyConnector.from_url(str(proxy))
                else:
                    connector = TCPConnector(**self._conn_kwargs)
                self._session = ClientSession(
                    connector=connector,
                    timeout=self._timeout,
                    json_serialize=json.dumps,
                )

        return self._session

    def __init__(
            self, *, limit: int = 30, timeout: float = 10,
            proxy: Optional[StrOrURL] = None, trust_env: bool = False
    ) -> None:
        self._session = None
        self._conn_kwargs = {'limit': limit}
        self._timeout = ClientTimeout(timeout)

        self._proxies = [URL(str(proxy))] if proxy is not None else []
        if trust_env:
            self._proxies.extend(proxies_from_env())
        self._proxies = list(set(self._proxies))

    async def __aenter__(self) -> Self:
        await self._init_session()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self._session.close()

    def __del__(self) -> None:
        if self._session and not self._session.closed:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()

            loop.run_until_complete(self._session.close())

    async def start(self) -> Self:
        await self._init_session()
        return self

    async def close(self) -> None:
        async with self._aio_lock:
            if self._session and not self._session.closed:
                await self._session.close()

    async def _request(
            self,
            method: RequestMethod,
            url: StrOrURL, *,
            params: Optional[Mapping[str, Any]] = None,
            headers: Optional[LooseHeaders] = None,
            data: Any = None,
    ) -> "ClientResponse":
        proxy = None
        for p in self._proxies:
            if p.scheme in ['http', 'ws']:
                proxy = p
                break
        param = dict()
        for key, value in (params or {}).items():
            if value in [None, {}, []]:
                continue
            if isinstance(value, Enum):
                value = value.value
            if isinstance(value, (str, int, float)):
                param.update({key: value})
            if isinstance(value, list):
                if key in ["sizes", "types"]:
                    param.update({key: ','.join(map(str, value))})
                elif key in ["ids"]:
                    param.update({
                        f"{key}[]": ",".join([str(pid) for pid in value])
                    })
                elif key in ["viewed"]:
                    [
                        param.update({f"{key}[{k}]": value[k]})
                        for k in range(len(value))
                    ]
                else:
                    param.update({f"{key}[]": ','.join(map(str, value))})
            if isinstance(value, bool):
                param.update({key: str(value).lower()})
        return await (await self._init_session()).request(
            method, url, params=param, headers=headers, data=data,
            proxy=proxy
        )

    async def get(
            self, url: StrOrURL, *, params: Optional[Mapping[str, Any]] = None,
            headers: Optional[LooseHeaders] = None, data: Any = None,
    ) -> "ClientResponse":
        return await self._request(
            'GET', url, params=params, headers=headers, data=data
        )

    async def post(
            self, url: StrOrURL, *, params: Optional[Mapping[str, Any]] = None,
            headers: Optional[LooseHeaders] = None, data: Any = None,
    ) -> "ClientResponse":
        return await self._request(
            'POST', url, params=params, headers=headers, data=data
        )

    async def head(
            self, url: StrOrURL, *, params: Optional[Mapping[str, Any]] = None,
            headers: Optional[LooseHeaders] = None, data: Any = None,
    ) -> "ClientResponse":
        return await self._request(
            'HEAD', url, params=params, headers=headers, data=data
        )

    async def put(
            self, url: StrOrURL, *, params: Optional[Mapping[str, Any]] = None,
            headers: Optional[LooseHeaders] = None, data: Any = None,
    ) -> "ClientResponse":
        return await self._request(
            "PUT", url, params=params, headers=headers, data=data
        )

    async def patch(
            self, url: StrOrURL, *, params: Optional[Mapping[str, Any]] = None,
            headers: Optional[LooseHeaders] = None, data: Any = None,
    ) -> "ClientResponse":
        return await self._request(
            "PATCH", url, params=params, headers=headers, data=data
        )

    async def delete(
            self, url: StrOrURL, *, params: Optional[Mapping[str, Any]] = None,
            headers: Optional[LooseHeaders] = None, data: Any = None,
    ) -> "ClientResponse":
        return await self._request(
            "DELETE", url, params=params, headers=headers, data=data
        )



from datetime import datetime
from enum import Enum
from typing import (
    List,
    Optional,
)

from pydantic import (
    EmailStr,
    HttpUrl,
)

from async_pixiv.model._base import PixivModel


class Artwork(PixivModel):
    class PageUrl(PixivModel):
        original_image_url: HttpUrl

    class MetaPage(PixivModel):
        image_urls: "ImageUrl"

    id: int
    title: str
    type: "ArtworkType"
    img_urls: List["ImageUrl"]
    caption: str
    restrict: int
    user: "User"
    tags: List["Tag"]
    tools: List[str]
    create_data: datetime
    page_count: int
    width: int
    height: int
    sanity_level: int
    x_restrict: int
    # series: Optional
    meta_single_page: PageUrl
    meta_pages: List[MetaPage]
    total_view: int
    total_bookmarks: int
    is_bookmarked: bool
    visible: bool
    is_muted: bool
    total_comments: int
    comment_access_control: int


class UgoiraMetadata(PixivModel):
    class ZipUrls(PixivModel):
        medium: HttpUrl

    class Frame(PixivModel):
        file: str
        delay: int

    zip_urls: ZipUrls
    frames: List[Frame]


class User(PixivModel):
    id: int
    name: str
    account: str
    profile_image_urls: "ProfileImg"
    is_followed: Optional[bool]
    is_premium: Optional[bool]
    is_mail_authorized: Optional[bool]
    x_restrict: Optional[int]
    mail_address: Optional[EmailStr]


class Tag(PixivModel):
    name: str
    translated_name: Optional[str]


class Comment(PixivModel):
    id: int
    comment: str
    date: datetime
    user: "User"
    parent_comment: Optional["Comment"]


class NovelSeries(PixivModel):
    id: int
    title: str


class Novel(PixivModel):
    id: int
    title: str
    caption: str
    restrict: int
    x_restrict: int
    is_original: bool
    image_urls: "ImageUrl"
    create_date: datetime
    tags: List["Tag"]
    page_count: int
    text_length: int
    user: "User"
    series: Optional["NovelSeries"]
    is_bookmarked: bool
    total_bookmarks: int
    total_view: int
    visible: bool
    total_comments: int
    is_muted: bool
    # noinspection SpellCheckingInspection
    is_mypixiv_only: bool
    is_x_restricted: bool
    comment_access_control: int


class ProfileImg(PixivModel):
    medium: Optional[HttpUrl]
    px_16x16: Optional[HttpUrl]
    px_50x50: Optional[HttpUrl]
    px_170x170: Optional[HttpUrl]


class ImageUrl(PixivModel):
    square_medium: HttpUrl
    medium: HttpUrl
    large: HttpUrl
    original: Optional[HttpUrl]


class ArtworkType(Enum):
    illust = 'illust'
    ugoira = 'ugoira'
    manga = 'manga'


class Profile(PixivModel):
    webpage: HttpUrl
    gender: str
    birth: str
    birth_day: str
    birth_year: int
    region: str
    address_id: int
    country_code: str
    job: str
    job_id: int
    total_follow_users: int
    # noinspection SpellCheckingInspection
    total_mypixiv_users: int
    total_illusts: int
    total_manga: int
    total_novels: int
    total_illust_bookmarks_public: int
    total_illust_series: int
    total_novel_series: int
    background_image_url: HttpUrl
    twitter_account: str
    twitter_url: HttpUrl
    pawoo_url: Optional[HttpUrl]
    is_premium: bool
    is_using_custom_profile_image: bool


class ProfilePublicity(PixivModel):
    class Choice(Enum):
        public = 'public'

    gender: Choice
    region: Choice
    birth_day: Choice
    birth_year: Choice
    job: Choice
    pawoo: Choice


class WorkSpace(PixivModel):
    pc: Optional[str]
    monitor: Optional[str]
    tool: Optional[str]
    scanner: Optional[str]
    tablet: Optional[str]
    mouse: Optional[str]
    printer: Optional[str]
    desktop: Optional[str]
    music: Optional[str]
    desk: Optional[str]
    chair: Optional[str]
    comment: Optional[str]
    workspace_image_url: Optional[HttpUrl]



from pydantic import (
    BaseConfig as PydanticBaseConfig,
    BaseModel as PydanticBaseModel,
)

__all__ = ['PixivModel']


class PixivModel(PydanticBaseModel):
    def __init__(self, **kwargs):
        self.__class__.update_forward_refs()
        super(PixivModel, self).__init__(**kwargs)

    class Config(PydanticBaseConfig):
        try:
            import ujson as json
        except ImportError:
            import json
        json_dumps = json.dumps
        json_loads = json.loads
        validate_all = True
        validate_assignment = True



from typing import (
    Callable,
    List,
    TypeVar,
)

from yarl import URL

__all__ = ['proxies_from_env']

T_Wrapped = TypeVar("T_Wrapped", bound=Callable)


# noinspection PyProtectedMember
def proxies_from_env() -> List[URL]:
    from urllib.request import getproxies
    return [
        URL(v)
        for k, v in getproxies().items()
        if k in ("https", "http", "socks4", "socks5", "ws", "wss")
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['async_pixiv', 'async_pixiv.client', 'async_pixiv.model', 'async_pixiv.utils']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles>=0.8.0,<0.9.0',
 'aiohttp-socks>=0.7.1,<0.8.0',
 'aiohttp>=3.8.1,<4.0.0',
 'arko-wrapper>=0.2.0,<0.3.0',
 'pydantic[email]>=1.9.1,<2.0.0']

extras_require = \
{'extra': ['regex>=2022.7.25,<2023.0.0', 'ujson>=5.4.0,<6.0.0'],
 'extra:sys_platform != "win32"': ['uvloop>=0.16.0,<0.17.0'],
 'full': ['regex>=2022.7.25,<2023.0.0',
          'ujson>=5.4.0,<6.0.0',
          'playwright>=1.24.0,<2.0.0'],
 'full:sys_platform != "win32"': ['uvloop>=0.16.0,<0.17.0'],
 'playwright': ['playwright>=1.24.0,<2.0.0'],
 'regex': ['regex>=2022.7.25,<2023.0.0'],
 'ujson': ['ujson>=5.4.0,<6.0.0']}

setup_kwargs = {
    'name': 'async-pixiv',
    'version': '0.0.6',
    'description': 'Async Pixiv API',
    'long_description': '# async-pixiv\nAsync Pixiv API\n\n\n## Building...',
    'author': 'Arko',
    'author_email': 'arko.space.cc@gmail.com',
    'maintainer': 'Karako',
    'maintainer_email': 'karakohear@gmail.com',
    'url': 'https://github.com/ArkoClub/async-pixiv',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from contextlib import contextmanager


@contextmanager
def no_warning() -> None:
    import warnings
    from copy import deepcopy
    filters = deepcopy(warnings.filters)
    try:
        warnings.filterwarnings("ignore")
        yield
    finally:
        warnings.resetwarnings()
        for f in filters:
            # noinspection PyUnresolvedReferences
            warnings.filters.append(filters)



from abc import ABC
from datetime import date
from enum import Enum as BaseEnum
from typing import (
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

from typing_extensions import Literal
from yarl import URL

from async_pixiv.model.result import (
    IllustCommentResult,
    IllustDetailResult,
    IllustRelatedResult,
    IllustSearchResult,
    UserBookmarksIllustsResult,
    UserDetailResult,
    UserIllustsResult,
    UserRelatedResult,
    UserSearchResult,
)

if TYPE_CHECKING:
    from ._client import PixivClient

__all__ = [
    'SearchShort', 'SearchDuration', 'SearchFilter',
    'SectionType',
    'USER', 'ILLUST'
]

API_HOST = URL("https://app-api.pixiv.net")
V1_API = API_HOST / 'v1'
V2_API = API_HOST / 'v2'


class Enum(BaseEnum):
    def __str__(self) -> str:
        # noinspection PyTypeChecker
        return self.value


class SearchShort(Enum):
    date_desc = 'date_desc'
    date_asc = 'date_asc'
    popular_desc = 'popular_desc'
    popular_asc = 'popular_asc'


class SearchDuration(Enum):
    day = 'within_last_day'
    week = 'within_last_week'
    month = 'within_last_month'
    year = 'within_last_year'


class SearchFilter(Enum):
    android = 'for_android'
    ios = 'for_ios'


# noinspection PyShadowingBuiltins
class _Section(ABC):
    _client: "PixivClient"
    _type: str = ''

    @property
    def type(self) -> str:
        return self._type

    def __init__(self, client: "PixivClient") -> None:
        self._client = client

    async def search(
            self,
            word: str, *,
            sort: Union[
                Literal['date_desc', 'date_asc', 'popular_desc', 'popular_asc'],
                SearchShort
            ] = SearchShort.date_desc,
            duration: Optional[Union[
                Literal[
                    'within_last_day',
                    'within_last_week',
                    'within_last_month',
                    'within_last_year'
                ], SearchDuration
            ]] = None,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: Optional[int] = None, **kwargs
    ) -> Dict:
        request = await self._client.get(
            V1_API / f'search/{self._type}',
            params={
                'word': word, 'sort': sort, 'duration': duration,
                'filter': filter, 'offset': offset, **kwargs
            }
        )
        return await request.json()

    async def detail(
            self, id: Optional[int] = None, *, filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios
    ) -> Dict:
        request = await self._client.get(
            V1_API / f'{self._type}/detail',
            params={f'{self._type}_id': id, 'filter': filter}
        )
        return await request.json()


SectionType = TypeVar('SectionType', bound=_Section)


class IllustType(Enum):
    illust = 'illust'
    manga = 'manga'
    ugoira = 'ugoira'
    novel = 'novel'


# noinspection PyShadowingBuiltins
class USER(_Section):
    _type = 'user'

    async def search(
            self,
            word: str, *,
            sort: Union[
                Literal['date_desc', 'date_asc', 'popular_desc', 'popular_asc'],
                SearchShort
            ] = SearchShort.date_desc,
            duration: Optional[Union[
                Literal[
                    'within_last_day',
                    'within_last_week',
                    'within_last_month',
                    'within_last_year'
                ], SearchDuration
            ]] = None,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: Optional[int] = None, **kwargs
    ) -> UserSearchResult:
        data = await super(USER, self).search(
            word=word, sort=sort, duration=duration, filter=filter,
            offset=offset, **kwargs
        )
        return UserSearchResult.parse_obj(data)

    async def detail(
            self, id: Optional[int] = None, *, filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios
    ) -> UserDetailResult:
        if id is None:
            id = self._client.account.id
        data = await super(USER, self).detail(id=id, filter=filter)
        return UserDetailResult.parse_obj(data)

    async def illusts(
            self, id: Optional[int] = None, *,
            type: Union[
                Literal['illust', 'manga', 'novel', 'ugoira'], IllustType
            ] = IllustType.illust,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: Optional[int] = None
    ) -> UserIllustsResult:
        if id is None:
            id = self._client.account.id
        data = await (await self._client.get(
            V1_API / "user/illusts",
            params={
                'user_id': id, 'type': type, 'filter': filter, 'offset': offset
            }
        )).json()
        return UserIllustsResult.parse_obj(data)

    async def bookmarks(
            self, id: Optional[int] = None, *,
            tag: Optional[str] = None,
            max_bookmark_id: Optional[int] = None,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
    ) -> UserBookmarksIllustsResult:
        if id is None:
            id = self._client.account.id
        data = await (await self._client.get(
            V1_API / "user/bookmarks/illust",
            params={
                'user_id': id, 'filter': filter, 'tag': tag,
                'max_bookmark_id': max_bookmark_id, 'restrict': 'public'
            }
        )).json()
        return UserBookmarksIllustsResult.parse_obj(data)

    async def related(
            self, id: Optional[int] = None, *,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: int = None,
    ) -> UserRelatedResult:
        if id is None:
            id = self._client.account.id
        data = await (await self._client.get(
            V1_API / "user/related",
            params={
                'seed_user_id': id, 'filter': filter, 'offset': offset
            }
        )).json()
        return UserRelatedResult.parse_obj(data)


# noinspection PyShadowingBuiltins
class ILLUST(_Section):
    _type = 'illust'

    async def follow(
            self, *, offset: Optional[int] = None
    ) -> IllustSearchResult:
        data = await (await self._client.get(
            V2_API / "illust/follow",
            params={'restrict': 'public', 'offset': offset}
        )).json()
        return IllustSearchResult.parse_obj(data)

    async def search(
            self,
            word: str, *,
            sort: Union[
                Literal['date_desc', 'date_asc', 'popular_desc', 'popular_asc'],
                SearchShort
            ] = SearchShort.date_desc,
            duration: Optional[Union[
                Literal[
                    'within_last_day',
                    'within_last_week',
                    'within_last_month',
                    'within_last_year'
                ], SearchDuration
            ]] = None,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: Optional[int] = None,
            min_bookmarks: Optional[int] = None,
            max_bookmarks: Optional[int] = None,
            start_date: Optional[date] = None,
            end_date: Optional[date] = None,
            **kwargs
    ) -> IllustSearchResult:
        if start_date is not None:
            start_date = start_date.strftime('%Y-%m-%d')
        if end_date is not None:
            end_date = end_date.strftime('%Y-%m-%d')
        data = await super(ILLUST, self).search(
            word=word, sort=sort, duration=duration, filter=filter,
            offset=offset, bookmark_num_min=min_bookmarks,
            bookmark_num_max=max_bookmarks,
            start_date=start_date, end_date=end_date,
        )
        return IllustSearchResult.parse_obj(data)

    async def detail(
            self, id: Optional[int] = None, *,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios
    ) -> IllustDetailResult:
        data = await super(ILLUST, self).detail(id, filter=filter)
        return IllustDetailResult.parse_obj(data)

    async def comments(
            self, id: int, *, offset: Optional[int] = None
    ) -> IllustCommentResult:
        data = await (await self._client.get(
            V1_API / "illust/comments", params={
                'illust_id': id, "offset": offset
            }
        )).json()
        return IllustCommentResult.parse_obj(data)

    async def related(
            self, id: int, *,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: Optional[int] = None,
            seed_ids: Optional[Union[List[int], int]] = None
    ) -> IllustRelatedResult:
        data = await (await self._client.get(
            V2_API / "illust/related", params={
                'illust_id': id, "offset": offset, 'filter': filter,
                'seed_illust_ids': (
                    [seed_ids] if seed_ids is not list else seed_ids
                )
            }
        )).json()
        return IllustRelatedResult.parse_obj(data)

    async def recommended(
            self, *, with_auth: bool = True, type: Union[
                Literal['illust', 'manga', 'novel', 'ugoira'], IllustType
            ] = IllustType.illust,
            include_ranking_label: bool = True,
            include_ranking_illusts: Optional[bool] = None,
            offset: Optional[int] = None,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            max_bookmark_id_for_recommend: Optional[int] = None,
            min_bookmark_id_for_recent_illust: Optional[int] = None,
            bookmark_illust_ids: Optional[Union[List[int], int]] = None,
            include_privacy_policy: Optional[Union[
                str, List[Union[int, str]]
            ]] = None,
            viewed: Optional[List[int]] = None
    ):
        data = await (await self._client.get(
            V1_API / (
                "illust/recommended"
                if with_auth else
                "illust/recommended-nologin"
            ), params={
                'content_type': type,
                'include_ranking_label': include_ranking_label,
                'include_ranking_illusts': include_ranking_illusts,
                'offset': offset, 'filter': filter, 'viewed': viewed,
                'max_bookmark_id_for_recommend': max_bookmark_id_for_recommend,
                'min_bookmark_id_for_recent_illust':
                    min_bookmark_id_for_recent_illust,
                'bookmark_illust_ids': bookmark_illust_ids,
                'include_privacy_policy': include_privacy_policy,
            }
        )).json()
        breakpoint()



from typing import Optional

from pydantic import HttpUrl

from async_pixiv.model._base import PixivModel


class ImageUrl(PixivModel):
    square_medium: Optional[HttpUrl]
    medium: Optional[HttpUrl]
    large: Optional[HttpUrl]
    original: Optional[HttpUrl]
